
/********************************************
 *  mcmap.c by Guillem Pratx
 *
 *  Runs a Map task consisting of simulating a given number of photons
 *  using the random seed read from standard input. After completion,
 *  the program writes the measured photon absorption to standard
 *  output, to be transfered to a Reduce task.
 *
 *  An article describes the use of this program in a cloud computing
 *  environment:
 *    G. Pratx & L. Xing, "Monte Carlo Simulation of Photon Migration 
 *    in a Cloud Computing Environment with MapReduce", 
 *    J. Biomed. Opt. in press. 
 * 
 *  Based on the MC321 package, which was developped by
 *  Steven L. Jacques based on prior collaborative work with Lihong
 *  Wang, Scott Prahl, and Marleen Keijzer.
 *  http://omlc.ogi.edu/classroom/ece532/class4/ssmc/index.html
 *
 *  A published report illustrates use of the MC321 program:
 *    S. L. Jacques: "Light distributions from point, line, and plane 
 *    sources for photochemical reactions and fluorescence in turbid 
 *    biological tissues," Photochem. Photobiol. 67:23-32, 1998. 
 **********/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define	PI          3.1415926
#define ALIVE       1   		/* if photon not yet terminated */
#define DEAD        0    		/* if photon is to be terminated */
#define THRESHOLD   0.01		/* used in roulette */
#define CHANCE      0.1  		/* used in roulette */
#define ONE_MINUS_COSZERO 1.0E-12
     /* If 1-cos(theta) <= ONE_MINUS_COSZERO, fabs(theta) <= 1e-6 rad. */
     /* If 1+cos(theta) <= ONE_MINUS_COSZERO, fabs(PI-theta) <= 1e-6 rad. */
#define SIGN(x)           ((x)>=0 ? 1:-1)
#define InitRandomGen    (double) RandomGen(0, 1, NULL)
     /* Initializes the seed for the random number generator. */     
#define RandomNum        (double) RandomGen(1, 0, NULL)
     /* Calls for a random number from the randum number generator. */

/* DECLARE FUNCTION */
double RandomGen(char Type, long Seed, long *Status);  
     /* Random number generator */

/* Converts a little-endian (LE) unsigned integer into the big-endian (BE)
format (and vice-versa) */ 
unsigned int ByteSwap( unsigned int n)
{
return	( (( n << 24) & 0xFF000000) |	\
	  (( n <<  8) & 0x00FF0000) |	\
	  (( n >>  8) & 0x0000FF00) |	\
	  (( n >> 24) & 0x000000FF) );
}

/* Casts a float as an unsigned integer */
unsigned int float_as_int (float f)
{
   return (*((int*)&f));
}

/* Casts an unsigned integer as a float */
float int_as_float (unsigned int i)
{
   return (*((float*)&i));
}

/*****************************************
[1] To send a uint32 integer to Hadoop:
unsigned int a, a_;
a_ = ByteSwap( a );
fwrite( &a_, sizeof(unsigned int), 1, stdout );

[2] To read a uint32 from Hadoop:
fread( &a_, sizeof(unsigned int), 1, stdin);
a = ByteSwap( a_ );

[3] To send a float to Hadoop:
float f;
unsigned int f_;
f_ = ByteSwap( float_as_int(f) );
fwrite( &f_, sizeof(unsigned int), 1, stdout );

[4] To read a float from Hadoop:
fread( &f_, sizeof(unsigned int), 1, stdin );
f = int_as_float(ByteSwap( f_ ));
**********************************************/

int main(int argc, const char* argv[] ) {

/* Propagation parameters */
float	x, y, z;      /* photon position */
float	ux, uy, uz;   /* photon trajectory as cosines */
float  uxx, uyy, uzz; /* temporary values used during SPIN */
float	s;            /* step sizes. s = -log(RND)/mus [cm] */
float	costheta;     /* cos(theta) */
float   sintheta;     /* sin(theta) */
float   cospsi;       /* cos(psi) */
float   sinpsi;       /* sin(psi) */
float	psi;          /* azimuthal angle */
float	i_photon;     /* current photon */
float	W;            /* photon weight */
float	absorb;       /* weighted deposited in a step due to absorption */
short   photon_status; /* flag = ALIVE=1 or DEAD=0 */

/* other variables */
float	mua;          /* absorption coefficient [cm^-1] */
float	mus;          /* scattering coefficient [cm^-1] */
float	g;            /* anisotropy [-] */
float	albedo;       /* albedo of tissue */
float	Nphotons;     /* number of photons in simulation per random seed */
float	rs;           /* phantom radius in cm */
float	r;            /* radial position */
float   dx,dy;        /* size of the histogram bins, in cm */
int	    ix,iy;        /* histogram indices */
int     i;            /* Loop index */
int     rflag = 1;

/* Variable convention: The _ suffix indicates a big-endian variable */
unsigned int key, key_; /* MapReduce key [little-endian (LE) and big_endian (BE)] */
float   *value;         /* MapReduce value (LE); pointer towards an array of floats */
unsigned int *value_;   /* MapReduce value (BE); pointer towards an array of floats */
long    seed;           /* Random seed used to initialize the random number generator */

/* dummy variables */
double  rnd;            /* assigned random value 0-1 */
float	temp;           /* dummy variables */

int     N;              /* Number of histogram bins in output*/
unsigned int N2, N2_;   /* size of histogram output (LE and BE).*/

/**** INPUT
   Input the optical properties
   Input the bin and array sizes 
   Input the number of photons
*****/

rs       = 2.5;
Nphotons = 10000;
N        = 256;
N2       = N*N;           /* scoring array size */
N2_      = ByteSwap(N2);  /* conversion from LE to BE */
mus      = 7.8;
mua      = 0.0025;
g        = 0.9;  
dx       = 2*rs/N;
dy       = 2*rs/N;
albedo   = mus/(mus + mua);

/* Allocate memory for storing histogram */
value = (float*) calloc( N2, sizeof(float) );
value_ = (unsigned int*) calloc( N2, sizeof(float) );

/**** Read seeds from standard input *****/
rflag = fscanf(stdin,"%ld",&seed);

/* Repeat while seeds are available */
while (rflag==1)
{

i_photon = 0;

/* Use seed to initialize the RNG*/
RandomGen(0, seed, NULL);

/* Read next seed from stdin, if available */
rflag = fscanf(stdin,"%ld",&seed);
   
/**** RUN
   Launch N photons, initializing each one before progation.
*****/
do {


/**** LAUNCH 
   Initialize photon position and trajectory.
   Implements an isotropic point source.
*****/
i_photon += 1;              /* increment photon count */
W = 1.0;                    /* set photon weight to one */
photon_status = ALIVE;      /* Launch an ALIVE photon */

x = -rs*0.99;               /* Set photon position to origin. */
y = 0;
z = 0;

/* Randomly set photon trajectory to yield an isotropic source. */
ux = 1;
uy = 0;
uz = 0;


/* HOP_DROP_SPIN_CHECK
   Propagate one photon until it dies as determined by ROULETTE.
*******/
do {


/**** HOP
   Take step to new position
   s = stepsize
   ux, uy, uz are cosines of current photon trajectory
*****/
  while ((rnd = RandomNum) <= 0.0);   /* yields 0 < rnd <= 1 */
  s = -log(rnd)/(mua + mus);          /* Step size.  Note: log() is base e */
  x += s * ux;                        /* Update positions. */
  y += s * uy;
  z += s * uz;


/**** DROP
   Drop photon weight (W) into local bin.
*****/
   absorb = W*(1 - albedo);      /* photon weight absorbed at this step */
   W -= absorb;                  /* decrement WEIGHT by amount absorbed */
   
   /* cylindrical */
   r = sqrt(x*x + y*y);          /* current cylindrical radial position */
   if (r<rs) {                   /* Check if photon still inside phantom */
      ix = (x+rs)/dx;            /* Compute histogram coordinates */
      iy = (y+rs)/dy;
      if (ix>=0 && ix < N && iy>=0 && iy < N)
        value[ix+N*iy] += absorb; /* DROP absorbed weight into bin */
   }
   else
      photon_status = DEAD;      /* Kill photon if outside phantom */
   
/**** SPIN 
   Scatter photon into new trajectory defined by theta and psi.
   Theta is specified by cos(theta), which is determined 
   based on the Henyey-Greenstein scattering function.
   Convert theta and psi into cosines ux, uy, uz. 
*****/
  /* Sample for costheta */
  rnd = RandomNum;
     if (g == 0.0)
        costheta = 2.0*rnd - 1.0;
     else {
        temp = (1.0 - g*g)/(1.0 - g + 2*g*rnd);
        costheta = (1.0 + g*g - temp*temp)/(2.0*g);
        }
  if (fabs(costheta)>1)                     /* Make sure that 1-cotheta*costheta>=0 */
    costheta = 1; 
  sintheta = sqrt(1.0 - costheta*costheta); /* sqrt() is faster than sin(). */

  /* Sample psi. */
  psi = 2.0*PI*RandomNum;
  cospsi = cos(psi);
  if (fabs(cospsi)>1)                       /* Make sure that 1-cotheta*costheta>=0 */
    cospsi = 1;
  if (psi < PI)
    sinpsi = sqrt(1.0 - cospsi*cospsi);     /* sqrt() is faster than sin(). */
  else
    sinpsi = -sqrt(1.0 - cospsi*cospsi);

  /* New trajectory. */
  if (1 - fabs(uz) <= ONE_MINUS_COSZERO) {  /* close to perpendicular. */
    uxx = sintheta * cospsi;
    uyy = sintheta * sinpsi;
    uzz = costheta * SIGN(uz);              /* SIGN() is faster than division. */
    } 
  else {					/* usually use this option */
    temp = sqrt(1.0 - uz * uz);
    uxx = sintheta * (ux * uz * cospsi - uy * sinpsi) / temp + ux * costheta;
    uyy = sintheta * (uy * uz * cospsi + ux * sinpsi) / temp + uy * costheta;
    uzz = -sintheta * cospsi * temp + uz * costheta;
    }
    
  /* Update trajectory */
  ux = uxx;
  uy = uyy;
  uz = uzz;


/**** CHECK ROULETTE 
   If photon weight below THRESHOLD, then terminate photon using Roulette technique.
   Photon has CHANCE probability of having its weight increased by factor of 1/CHANCE,
   and 1-CHANCE probability of terminating.
*****/
  if (W < THRESHOLD) {
   if (RandomNum <= CHANCE)
      W /= CHANCE;
   else photon_status = DEAD;
   }


} /* end STEP_CHECK_HOP_SPIN */
while (photon_status == ALIVE);

  /* If photon dead, then launch new photon. */
} /* end RUN */
while (i_photon < Nphotons);

/* End of run for current random seed */
}

/* write photon absorption histogram to stdout*/

key = 1;                                 /* key is an integer arbitrarily set to "1" */
key_ = ByteSwap(key);                    /* key is converted to BE */

/* A TypedBytes sequence starts with a byte code. Here, the Key is an integer (code: 3)
http://hadoop.apache.org/mapreduce/docs/r0.21.0/api/org/apache/hadoop/typedbytes/package-summary.html
*/
fputc( 3, stdout );                       /* write byte code*/    
fwrite( &key_, sizeof(int), 1, stdout );  /* write integer key */

/* The Value is an array of floats. Byte code 8 indicates a vector, followed by a 32-bit integer
length, followed by as many TypedBytes sequences as indicated by the length */
fputc( 8, stdout );                       /* write byte code*/
fwrite( &N2_, sizeof(int), 1, stdout );   /* write array length (N2 = N*N). Note that N2_ is BE */

/* Write out individual array element, consisting of a byte code (5 for a float)
followed by a BE-format float*/
for (i=0; i<N2; i++) {
  value_[i] = ByteSwap(float_as_int(value[i]));          /* convert float from LE to BE */
  fputc( 5, stdout );                                    /* write byte code */
  fwrite( &value_[i], sizeof(unsigned int), 1, stdout ); /* write BE Value */
}

free(value);                               /* Free memory */
free(value_);

return 0;

} /* end of main */

 

/* SUBROUTINES */

/**************************************************************************
 *	RandomGen
 *      A random number generator that generates uniformly
 *      distributed random numbers between 0 and 1 inclusive.
 *      The algorithm is based on:
 *      W.H. Press, S.A. Teukolsky, W.T. Vetterling, and B.P.
 *      Flannery, "Numerical Recipes in C," Cambridge University
 *      Press, 2nd edition, (1992).
 *      and
 *      D.E. Knuth, "Seminumerical Algorithms," 2nd edition, vol. 2
 *      of "The Art of Computer Programming", Addison-Wesley, (1981).
 *
 *      When Type is 0, sets Seed as the seed. Make sure 0<Seed<32000.
 *      When Type is 1, returns a random number.
 *      When Type is 2, gets the status of the generator.
 *      When Type is 3, restores the status of the generator.
 *
 *      The status of the generator is represented by Status[0..56].
 *
 *      Make sure you initialize the seed before you get random
 *      numbers.
 ****/
#define MBIG 1000000000
#define MSEED 161803398
#define MZ 0
#define FAC 1.0E-9

double RandomGen(char Type, long Seed, long *Status){
  static long i1, i2, ma[56];   /* ma[0] is not used. */
  long        mj, mk;
  short       i, ii;

  if (Type == 0) {              /* set seed. */
// Added by GP: clear state
    for (i=0;i<56;i++)
      ma[i] = 0;
    i1=0;    i2=0;
// GP
    mj = MSEED - (Seed < 0 ? -Seed : Seed);
    mj %= MBIG;
    ma[55] = mj;
    mk = 1;
    for (i = 1; i <= 54; i++) {
      ii = (21 * i) % 55;
      ma[ii] = mk;
      mk = mj - mk;
      if (mk < MZ)
        mk += MBIG;
      mj = ma[ii];
    }
    for (ii = 1; ii <= 4; ii++)
      for (i = 1; i <= 55; i++) {
        ma[i] -= ma[1 + (i + 30) % 55];
        if (ma[i] < MZ)
          ma[i] += MBIG;
      }
    i1 = 0;
    i2 = 31;
  } else if (Type == 1) {       /* get a number. */
    if (++i1 == 56)
      i1 = 1;
    if (++i2 == 56)
      i2 = 1;
    mj = ma[i1] - ma[i2];
    if (mj < MZ)
      mj += MBIG;
    ma[i1] = mj;
    return (mj * FAC);
  } else if (Type == 2) {       /* get status. */
    for (i = 0; i < 55; i++)
      Status[i] = ma[i + 1];
    Status[55] = i1;
    Status[56] = i2;
  } else if (Type == 3) {       /* restore status. */
    for (i = 0; i < 55; i++)
      ma[i + 1] = Status[i];
    i1 = Status[55];
    i2 = Status[56];
  } else
    puts("Wrong parameter to RandomGen().");
  return (0);
}
#undef MBIG
#undef MSEED
#undef MZ
#undef FAC

