/********************************************
 *  mcred.c by Guillem Pratx
 *
 *  Runs a Reduce task. The Reduce task accumulates the scoring arrays
 *  from many Map tasks.
 *
 *  An article describes the use of this program in a cloud computing
 *  environment:
 *    G. Pratx & L. Xing, "Monte Carlo Simulation of Photon Migration 
 *    in a Cloud Computing Environment with MapReduce", 
 *    J. Biomed. Opt. in press. 
 * 
 *  Based on the MC321 package, which was developped by
 *  Steven L. Jacques based on prior collaborative work with Lihong
 *  Wang, Scott Prahl, and Marleen Keijzer.
 *  http://omlc.ogi.edu/classroom/ece532/class4/ssmc/index.html
 *
 *  A published report illustrates use of the MC321 program:
 *    S. L. Jacques: "Light distributions from point, line, and plane 
 *    sources for photochemical reactions and fluorescence in turbid 
 *    biological tissues," Photochem. Photobiol. 67:23-32, 1998. 
 **********/

#include <stdio.h>
#include <stdlib.h>

/* Converts a little-endian (LE) unsigned integer into the big-endian (BE)
format (and vice-versa) */ 
unsigned int ByteSwap( unsigned int n )
{
return	( (( n << 24) & 0xFF000000) |	\
	  (( n <<  8) & 0x00FF0000) |	\
	  (( n >>  8) & 0x0000FF00) |	\
	  (( n >> 24) & 0x000000FF) );
}

/* Casts a float as an unsigned integer */
unsigned int float_as_int ( float f )
{
   return (*((int*)&f));
}

/* Casts an unsigned integer as a float */
float int_as_float ( unsigned int i )
{
   return (*((float*)&i));
}

float        *value,      /* scoring array, LE */
             *sum;        /* cumulative scoring array, LE */
unsigned int *value_,     /* scoring array, BE */
             *sum_,       /* cumulative scoring array, BE */
             key,         /* key, set to 1, LE */
             key_,        /* key, BE */
             l,           /* scoring array length, LE */
             l_;          /* scoring array length, BE */
int          r=1,         /* flag returned by fread */
             i;
char         c;


int main(int argc, const char* argv[] ) {

/* Read key, consisting of byte code (3) and integer */
c   = fgetc( stdin );
r   = fread( &key_, sizeof(int), 1, stdin );
key = ByteSwap( key_ );

/* Read scoring array: byte code (8), length, sequence of floats */
c = fgetc( stdin );
r = fread( &l_, sizeof(float), 1, stdin );
l = ByteSwap( l_ );

/* allocate scoring array (LE and BE) */
value = (float*) calloc( l, sizeof(float) );
value_ = (unsigned int*) calloc( l, sizeof(float) );

/* allocate cumulative scoring array (LE and BE) */
sum = (float*) calloc( l, sizeof(float) );
sum_ = (unsigned int*) calloc( l, sizeof(float) );

/* read scoring array from Map task via standard input, and convert to LE */
for ( i=0; i<l; i++ ) {
  c = fgetc( stdin );
  r = fread( &value_[i], sizeof(unsigned int), 1, stdin );
  value[i] = int_as_float( ByteSwap( value_[i] ) );
}

/* Repeat while new records can be read from stdin */
while ( r==1 ) {
  /* accumulate last read array into sum */
  for ( i=0; i<l; i++ )
    sum[i] += value[i];

  /* Read next key */
  c = fgetc( stdin );
  r = fread( &key_, sizeof(int), 1, stdin );
  key=ByteSwap(key_);

  /* Read next scoring array */
  c = fgetc( stdin );
  r = fread( &l_, sizeof(float), 1, stdin );
  l = ByteSwap( l_ );

  for ( i=0; i<l && r==1; i++ ) {
    c = fgetc( stdin );
    r = fread( &value_[i], sizeof(unsigned int), 1, stdin );
    value[i] = int_as_float( ByteSwap( value_[i] ) );
  }
}

/* output key */
key  = 1;
key_ = ByteSwap(key);
fputc( 3, stdout );
fwrite( &key_, sizeof(int), 1, stdout );

/* output cumulated array */
fputc( 8, stdout );
fwrite( &l_, sizeof(int), 1, stdout );

for ( i=0; i<l; i++ ) {
  sum_[i] = ByteSwap(float_as_int(sum[i]));
  fputc( 5, stdout );
  fwrite( &sum_[i], sizeof(unsigned int), 1, stdout );
}

return 0;
}


